from typing import List, Generator
from pathlib import Path
from collections import defaultdict

from python_mecab_ner.domain.mecab_exception import MecabDataReaderException
from python_mecab_ner.service.mecab_parser import MecabParser


class CategoryData:
    """
    large_cat, small_cat에 word 리스트로 저장
    """
    def __init__(self):
        self.large_cat = defaultdict(list)

    def add(self, small_cat, word):
        self.large_cat[small_cat].extend(word)


class DataUtility:

    """
    데이터 읽고 쓰기 유틸리티
    """

    @staticmethod
    def read_txt(path):
        """텍스트 파일 읽는 메소드"""
        with open(path, "r", encoding='utf-8-sig') as file:
            txt_list = file.read().splitlines()
            return txt_list

    @staticmethod
    def write_txt(path: str, txt_list: List, is_sort=False):
        """텍스트 파일 쓰는 메소드"""
        if is_sort:
            txt_list = sorted(list(txt_list), key=len, reverse=True)

        with open(path, "a", encoding='UTF8') as file:
            for idx, txt_item in enumerate(txt_list):
                data = str(txt_item) + "\n"
                file.write(data)


class MecabDataController:

    """
    데이터 읽고 쓰는 기능
    """

    FIRST_WORD = 0
    MECAB_WORD = 1
    ORIGIN_WORD = 0

    MECAB_DATA = "mecab_data"
    ITEM_BOUNDARY = ","
    CATEGORY_SPLITER = "_"
    FORMAT_SUFFIX = ".txt"
    SMALL_CAT_DIVIDER = "#"

    def __init__(self, ner_path: str = None, clear_mecab_dir=True):

        """
        :param ner_path: ner 데이터 경로
        :param clear_mecab_dir: 기존 메캅 디렉터리 삭제 여부
        """

        self.ner_path = ner_path or "./"
        self._clear_mecab_dir = clear_mecab_dir
        self._set_mecab_path(self.ner_path)

    def _set_mecab_path(self, ner_path: str) -> None:

        """
        메캅 데이터 경로 생성
        - 경로가 정상 디렉터리인지 확인
        :param ner_path: ner_path 입력 경로
        :param clear_dir: 경로 전부 삭제할 것인지 여부
        """

        if not Path(ner_path).is_dir():
            raise MecabDataReaderException("Please check if directory is proper")

        self.mecab_path = Path(__file__).parent.parent.joinpath("data", self.MECAB_DATA)

        if self._clear_mecab_dir:
            self._clear_dir()

        if not Path(self.mecab_path).exists():
            Path(self.mecab_path).mkdir()

    def _clear_dir(self):

        """
        메캅 데이터 저장 디렉터리 데이터 전부 삭제하는 메소드
        - ner_example인 경우 제외
        """

        try:
            for path_item in Path(self.mecab_path).iterdir():

                if path_item.stem.startswith("ner_example"): # 예시코드 삭제 방지 코드
                    continue

                Path(path_item).unlink()

        except FileNotFoundError:
            pass

    @classmethod
    def read_category(cls, datas: List) -> Generator:

        """
        데이터에서 헤더, 내용을 나눠서 반환하는 메소드
        SMALL_CAT로 시작되지 않을시 small_category 리스트로 반환
        """

        small_cat, *words = datas
        small_cat_words = []

        for word in words:
            if cls.SMALL_CAT_DIVIDER in word:
                yield small_cat, sorted(small_cat_words, key=len, reverse=True)
                small_cat = word
                small_cat_words = []
                continue

            if word == '':
                continue

            small_cat_words.append(word)

        yield small_cat, sorted(small_cat_words, key=len, reverse=True)

    @classmethod
    def gen_all_mecab_category_data(cls, path, use_mecab_parser=False) -> Generator:

        """
        경로에 있는 데이터 읽은 뒤 카테고리 데이터셋으로 반환
        :param path: 데이터 읽는 경로. mecab_data or ner_data경로 접근 가능
        :param use_mecab_parser: 메캅 파서 사용시 분해되서 반환.
        :return: 카테고리이름, 데이터 반환
        """

        for path_item in Path(path).iterdir():
            if Path(path_item).suffix != cls.FORMAT_SUFFIX:
                continue

            txt_data = DataUtility.read_txt(path_item)
            large_category = path_item.stem
            c_d = CategoryData()

            if not txt_data[cls.FIRST_WORD].startswith(cls.SMALL_CAT_DIVIDER):
                txt_data.insert(cls.FIRST_WORD, cls.SMALL_CAT_DIVIDER + path_item.stem)

            for small_cat_word in cls.read_category(txt_data):

                small_cat, words = small_cat_word

                if use_mecab_parser:
                    words = [(word, MecabParser(sentence=word).get_word_from_mecab_compound()) for word in words]

                c_d.add(small_cat, words)

            yield large_category, dict(c_d.large_cat)

    def write_category(self) -> None:

        """
        카테고리별로 데이터 저장하는 메소드
        ner_path에서 데이터 읽어서 mecab_path에 데이터 저장
        """

        for data_item in self.gen_all_mecab_category_data(path=self.ner_path, use_mecab_parser=True):
            category, content = data_item

            file_name = category + self.FORMAT_SUFFIX
            mecab_write_path = self.mecab_path.joinpath(file_name)

            if file_name.startswith("ner_example"): # 예시코드 실행시 추가 방지 코드
                continue

            for content_key_item in content.keys():
                mecab_write_list = [content_key_item,]
                mecab_write_list.extend([str(x[self.ORIGIN_WORD]) + self.ITEM_BOUNDARY + str(x[self.MECAB_WORD]) for x in
                             content[content_key_item]])
                DataUtility.write_txt(path=str(mecab_write_path), txt_list=mecab_write_list)





from collections import defaultdict


NODE_POS = 0
NODE_EXPRESSION = 1
MECAB_WORD_FEATURE = 1
IDX_TOKEN = 0

class MecabStorage:

    """
    메캅 형태소 분석 결과를 원래 문장으로 복원
    """

    def __init__(self):
        self.data = defaultdict(list)

    def _append(self, index, token):

        """
        인덱스
        :param index: 단어가 저장된 인덱스
        :param token: 단어
        """

        self.data[index].append(token)

    def _mecab_reverse(self):

        """
        메캅 결과에서 반환
        :return: 인덱스에 저장된 토큰 반환
        """

        reverse_sentence = list()
        for key in sorted(self.data):
            reverse_sentence.append("".join(self.data[key]))
        return reverse_sentence

    def restore_mecab_tokens(self, parsed_tokens):

        """
        복합어로 분해되지 않은 토큰들을 원래 문장으로 복원
        :param parsed_tokens: 파싱된 토큰들
        :return: 파싱된 토큰 합친 토큰
        """

        for parse_token_item in parsed_tokens:
            self._append(parse_token_item.space_token_idx, parse_token_item.word)
        return self._mecab_reverse()

    def reverse_compound_tokens(self, parse_compound_tokens):

        """
        복합어로 파싱된 결과를 space_token_idx을 이용해 원래대로 돌려주는 기능
        - 복합어가 아니라면 바로 붙임
        - 복합어는 앞에 복합어가 저장되어있는지 확인
        :param parse_compound_tokens: 복원할 토큰
        :return: 복원된 문장
        """

        tmp_word = None
        tmp_idx = None

        for parse_token_item in parse_compound_tokens:

            if parse_token_item[MECAB_WORD_FEATURE].type is None:
                self._append(parse_token_item[MECAB_WORD_FEATURE].space_token_idx, parse_token_item[MECAB_WORD_FEATURE].word)
                tmp_word = None
                continue

            if tmp_word == parse_token_item[MECAB_WORD_FEATURE].word and (
                    tmp_idx == parse_token_item[MECAB_WORD_FEATURE].space_token_idx):
                continue

            self._append(parse_token_item[MECAB_WORD_FEATURE].space_token_idx, parse_token_item[MECAB_WORD_FEATURE].word)
            tmp_word = parse_token_item[MECAB_WORD_FEATURE].word
            tmp_idx = parse_token_item[MECAB_WORD_FEATURE].space_token_idx

        return self._mecab_reverse()


from dataclasses import dataclass
from typing import Optional


@dataclass
class Category:
    large: str
    small: str


@dataclass
class MecabWordCategory:
    category: Category
    start_idx: Optional[int] = None
    end_idx: Optional[int] = None
    entity: Optional[str] = None


@dataclass
class MecabPatternData:
    category: Category
    dictionary_data: str
    pattern: str
    sentence: list
    min_meaning: int = 0
    parse_character: bool = False


@dataclass
class MecabWordFeature:
    word: str
    pos: str
    semantic: str
    has_jongseong: bool
    reading: str
    type : str
    start_pos: str
    end_pos: str
    expression: str
    space_token_idx: Optional[int] = None
    mecab_token_idx: Optional[int] = None
    mecab_token_compound_idx: Optional[int] = None


@dataclass
class MecabNerFeature:
    word: str
    pos: str
    start_idx: int
    end_idx: int
    category: Optional[Category] = None


@dataclass
class NerFeature:
    word: str
    pos: str
    category: Optional[Category] = None


import copy
from typing import List, Generator, Iterable
from pathlib import Path

from python_mecab_ner.service.mecab_parser import MecabParser
from python_mecab_ner.service.mecab_storage import MecabStorage
from python_mecab_ner.service.mecab_reader import MecabDataController
from python_mecab_ner.domain.mecab_domain import MecabWordCategory, Category, MecabPatternData, MecabNerFeature, NerFeature

MECAB_READING_WORD = 0
MECAB_FEATURE = 1
WORD_IDX = 0
INFER_FORWARD = 1
INFER_BACKWARD = 2
FULL_WORD = 1
START_IDX = 0
END_IDX = 1
EMPTY_WORD = 0


def find_patterns_idx(pattern, target_tokens: List, parse_character=False) -> List:

    """
    리스트에서 단어 패턴을 찾아서 반환하는 함수
    :param pattern: 찾고자 하는 패턴
    :param target_tokens: 찾아야 하는 토큰 리스트
    :param parse_character: 캐릭터 단위로 분해
    :return: 저장되어 있는 리스트 반환
    """

    if parse_character:
        pattern = list(pattern)

    if isinstance(pattern, str):
        pattern = pattern.split()

    tmp_save_list = []

    for i in range(len(target_tokens) - len(pattern) + 1):
        for j in range(len(pattern)):
            if target_tokens[i + j][MECAB_READING_WORD] != pattern[j]:
                break
        else:
            tmp_save_list.append((i, i+len(pattern)))

    return tmp_save_list



def gen_integrated_entity(blank_list: List) -> Generator:

    """
    1로 채워진 리스트의 시작점과 끝점을 반환
    """

    start_idx = None
    end_idx = None
    is_word_start = False
    for idx, item in enumerate(blank_list):
        if item == FULL_WORD: # 1이 채워진 단어에 대해서 수행
            end_idx = idx # 끝단어를 시작점으로 잡는다.

            if is_word_start is False: # 1로 채워진 단어의 시작이라면, 시작점을 현 인덱스로 저장한다.
                start_idx = idx
                is_word_start = True

            if idx != len(blank_list) - 1: # 인덱스가 끝이 아니라면, 다음 단어를 탐색한다.
                continue

        if (is_word_start is True) and (end_idx is not None):
            yield start_idx, end_idx
            start_idx = None
            end_idx = None
            is_word_start = False # 단어 초기화
            continue


def _get_default_mecab_large_cat(mecab_path):

    """
    지정한 카테고리 없을 시 메캅 경로를 기본 분석 경로로 지정 위한 메소드
    :param mecab_path: 메캅 분석된 데이터 저장 경로
    :return: 저장된 카테고리
    """

    for path_item in Path(mecab_path).iterdir():
        if path_item.suffix == ".txt":
            yield path_item.stem


def _prevent_compound_token(pattern_item: List, mc_p_l: List) -> None:
    """
    - 한 카테고리에서 같은 짧은 단어 추출하는 경우 방지. ex) 양념 치킨 추출 후 치킨 추출
    :param pattern_item: 찾는 패턴
    :param mc_p_l: 문장을 메캅으로 나눈 리스트
    """

    for pattern_idx_item in range(pattern_item[START_IDX], pattern_item[END_IDX], 1):
        mc_p_l[pattern_idx_item] = ("*", mc_p_l[pattern_idx_item][MECAB_FEATURE])


class MecabNer(MecabDataController):
    """
    엔티티 추출하는 클래스
    - 앞에 단어 품사를 통한 추론 기능
    - 메캅 형태소 분석 비교 및 형태소와 문자열 비교 기능
    """

    MIN_MEANING = 2
    NER_POS = "ner"
    ENTITY_POS_LIST = ["NNG", "NNP", "NNB", "NNBC", "NR", "NP", "XSN", "XR", "SL", "SH", "SN", "UNKNOWN"]
    INFER_ENTITY_POS_LIST = ["NNG", "NNP"]

    def __init__(self, ner_path: str = None, search_category: List = None, infer=True, clear_mecab_dir=True):

        """
        :param ner_path: ner 경로
        :param search_category: 찾고자 하는 카테고리
        :param infer: 추론 여부
        :param clear_mecab_dir: 기존 디렉터리 삭제 여부
        """

        super().__init__(ner_path=ner_path, clear_mecab_dir=clear_mecab_dir)
        self.search_category = search_category
        self.write_category()

        self.mecab_parsed_list = []

        if search_category is None:
            self.search_category = list(_get_default_mecab_large_cat(mecab_path= self.mecab_path))

        self.infer = infer

    def get_category_entity(self) -> Iterable[MecabWordCategory]:

        """
        문장에서 카테고리별로 엔티티 추출
        :param sentence: 엔티티를 찾고자 하는 문장
        :return: 추출 엔티티
        """

        for mecab_category_item in self.gen_all_mecab_category_data(path=self.mecab_path, use_mecab_parser=False):
            mecab_parsed_copied = copy.deepcopy(self.mecab_parsed_list)
            category, mecab_dictionary_data = mecab_category_item
            if category not in self.search_category:
                continue
            for small_category in mecab_dictionary_data.keys():

                for small_category_item in mecab_dictionary_data.get(small_category):

                    original_data, mecab_data = small_category_item.split(",")

                    category_data = Category(large=category, small=small_category)
                    yield from self._get_pattern(MecabPatternData(category=category_data, dictionary_data=original_data, pattern=mecab_data, sentence=mecab_parsed_copied))
                    yield from self._get_pattern(MecabPatternData(category=category_data, dictionary_data=original_data, pattern=original_data, sentence=mecab_parsed_copied, min_meaning=2, parse_character=True))

    def _get_pattern(self, m_p_d: MecabPatternData) -> Iterable[MecabWordCategory]:
        """
        문장에서 데이터 패턴을 찾는 기능
        - 엔티티의 마지막 단어는 명사형이여야 함. (이가 조사임에도 엔티티로 빠지는 경우 방지)
        :param m_p_d: 찾고자 하는 메캅 데이터 패턴 정보를 담고 있는 클래스
        :return: 문장에서 메캅 단어 정보를 담고 있는 클래스
        """
        space_token_contain_pattern = find_patterns_idx(m_p_d.pattern, m_p_d.sentence, m_p_d.parse_character)

        if (len(m_p_d.pattern) >= m_p_d.min_meaning) and space_token_contain_pattern:
            for pattern_item in space_token_contain_pattern:

                pattern_end_pos = m_p_d.sentence[pattern_item[END_IDX] - 1][MECAB_FEATURE].pos

                if pattern_end_pos not in self.ENTITY_POS_LIST and len(m_p_d.pattern) < self.MIN_MEANING:
                    continue

                _prevent_compound_token(pattern_item, m_p_d.sentence)

                yield MecabWordCategory(category=m_p_d.category,
                                        start_idx=pattern_item[START_IDX],
                                        end_idx=pattern_item[END_IDX], entity=m_p_d.dictionary_data)

    def infer_entity(self, mecab_category_item: MecabWordCategory) -> MecabWordCategory:

        """
        - mecab_category_item의 앞에 명사형이 올 경우 카테고리 아이템의 인덱스 값 변경
        :param mecab_category_item:  메캅 카테고리 아이템
        :return: 앞 단어 형태소에 따라 변경된 카테고리 아이템
        """

        end_point = 0
        if mecab_category_item.start_idx == 1:
            end_point = -1

        for idx_range in range(mecab_category_item.start_idx - 1, end_point, -1):
            if self.mecab_parsed_list[idx_range][MECAB_FEATURE].pos in self.INFER_ENTITY_POS_LIST:
                mecab_category_item.start_idx = self.mecab_parsed_list[idx_range][MECAB_FEATURE].mecab_token_compound_idx
                continue
            break
        return mecab_category_item


    def fill_entity_in_blank(self, mecab_entity_category_list: List) -> List:

        """
        문장에서 인덱스 값이 있는 경우 1로 채워서 반환
        :param mecab_entity_category_list:
        :return: 메캅 특성의 시작, 끝점이 있는 인덱스는 1로 채워서 반환
        """

        blank_list = [EMPTY_WORD] * len(self.mecab_parsed_list)

        for mecab_entity_category_item in mecab_entity_category_list:
            for i in range(mecab_entity_category_item.start_idx, mecab_entity_category_item.end_idx, 1):
                blank_list[i] = FULL_WORD

        return blank_list

    def gen_infer_mecab_ner_feature(self) -> Iterable[MecabNerFeature]:

        """
        추론 메캅 단어 특성 반환
        :return: 추론된 단어 정식 형태로로 분석하여 변환
        """

        if self.infer:
            mecab_entity_category_list = []
            for category_entity_item in self.get_category_entity():
                mecab_entity_category_list.append(self.infer_entity(category_entity_item))
        else:
            mecab_entity_category_list = list(self.get_category_entity())

        many_entity_index_list = self.fill_entity_in_blank(mecab_entity_category_list)

        for integrated_entity_item in gen_integrated_entity(many_entity_index_list):
            end_idx = integrated_entity_item[END_IDX] + 1
            start_idx = integrated_entity_item[START_IDX]
            mecab_parsed_token = self.mecab_parsed_list[start_idx:end_idx]
            restore_tokens = MecabStorage().reverse_compound_tokens(mecab_parsed_token)
            restore_sentence = " ".join(restore_tokens)
            for entity_category_item in mecab_entity_category_list:
                if entity_category_item.end_idx == end_idx:
                    small_category_replace = entity_category_item.category.small.replace(MecabDataController.SMALL_CAT_DIVIDER, "").strip()
                    yield MecabNerFeature(word=restore_sentence,
                                        pos=self.NER_POS,
                                        start_idx=start_idx,
                                        end_idx=end_idx,
                                        category=Category(large=entity_category_item.category.large, small=small_category_replace))


    def parse(self, sentence: str):
        """
        문장 분해 후 값 돌려주는 기능
        :param sentence: 입력 문장
        :return: 파싱돈 결과
        """

        self.mecab_parsed_list = list(MecabParser(sentence=sentence).gen_mecab_compound_token_feature())

        mecab_cat_list = list(self.gen_infer_mecab_ner_feature())
        cat_idx_list=[]
        [cat_idx_list.extend(list(range(x.start_idx, x.end_idx, 1))) for x in mecab_cat_list]

        parse_result = []
        for idx, mecab_parse_item in enumerate(self.mecab_parsed_list):

            for mecab_item in mecab_cat_list:
                if idx == mecab_item.end_idx:
                    parse_result.append((mecab_item.word,
                                         NerFeature(word=mecab_item.word,
                                                    pos=mecab_item.pos,
                                                    category=mecab_item.category)))
                    break
            if idx in cat_idx_list:
                continue

            parse_result.append((mecab_parse_item[MECAB_FEATURE].word, NerFeature(word=mecab_parse_item[MECAB_FEATURE].word,
                                                                                  pos=mecab_parse_item[MECAB_FEATURE].pos)))

        return parse_result

    def morphs(self, sentence: str):
        return [x[MECAB_READING_WORD] for x in self.parse(sentence=sentence)]

    def ners(self, sentence: str):
        result = [(x[MECAB_FEATURE].word, x[MECAB_FEATURE].category.large, x[MECAB_FEATURE].category.small) for x in self.parse(sentence=sentence) if x[MECAB_FEATURE].pos == "ner"]
        return result


from setuptools import setup, find_packages
# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name             = 'python-mecab-ner',
    version          = '1.0.2',
    description      = 'Test package for distribution',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author           = 'Youngchan',
    author_email     = 'youngchanchatbot@gmail.com',
    url              = 'https://github.com/YoungchanChang/python-mecab-ner',
    download_url     = '',
    install_requires = ['pybind11 ~= 2.0', "python-mecab-ko ~= 1.0.12"],
    keywords         = ['Text Processing', 'Text Processing :: Linguistic', 'NER', "mecab", "mecab-ko", "mecab-ner"],
    python_requires  = '>=3.7',
    zip_safe=False,
    classifiers      = [
        "Programming Language :: Python :: 3",
        'Programming Language :: Python :: 3 :: Only',
        "License :: OSI Approved :: MIT License",
        'Intended Audience :: Science/Research',
        'Natural Language :: Korean',
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    include_package_data=True,
    package_data={
        "": ["*.txt", "*.py"],
    }
)

